require("io")
require("lfs")
require("uuid")
local constants = require("shared.streaming.constants")
local http = require("socket.http")
local os = require "os"
local string = require "string"

module(..., package.seeall)

--////////////////////////////////////////////////////////////////////////////////////////////////////////
-- FUNCIONES PRIVADAS
--////////////////////////////////////////////////////////////////////////////////////////////////////////



local DIR_NAME = "\\inconcert\\web\\mediaServer\\"
local TEMPORAL_DIRECTORY_NAME = string.gsub(os.getenv("TEMP"), "\\", "\\\\")
local FILE_EXTENSION = {
	CALL = ".MP3",
	WAV = ".WAV",
	CHAT = ".XML",
	FAX = ".TIF"
}


local GetTemporalDirectory = function (typeName)
	return TEMPORAL_DIRECTORY_NAME .. DIR_NAME .. (typeName or "CALL")
end

local GetTemporalFileName = function (canCreateFile, typeName)
	if canCreateFile then
		local fileExtension = FILE_EXTENSION[typeName] or "CALL"
		local newFileName = uuid.new() .. fileExtension
		--return GetTemporalDirectory(typeName) .. "\\" .. newFileName
		return TEMPORAL_DIRECTORY_NAME .. "\\" .. newFileName
	end
	return GetTemporalDirectory(typeName) .. "\\defaultname"
end



--////////////////////////////////////////////////////////////////////////////////////////////////////////
--
local MakeParametersURL = function (parameters)
	if type(parameters) ~= "table" then
		return ""
	end
	local parametersString
	for property, value in pairs(parameters) do
		if (parametersString) then
			parametersString = parametersString .. "&"
		else
			parametersString = "?"
		end
		parametersString = parametersString .. property .. "=" .. value
	end
	return parametersString
end

-- ////////////////////////////////////////////////////////////////////////////////////////////////////////
-- Retorna la url base del repositorio, o sea, sin los parmetros
local GetRepositoryUrl = function (parameters)
	return "http://" .. parameters.repositoryIp .. ":" .. parameters.repositoryPort .. "/inconcert/apps/repository/get/"
end 

--[=[
Hace un request al repositorio y junta la respuesta en un array.
]=]
local function Request(parameters, credentials)
	local url = GetRepositoryUrl(parameters) .. MakeParametersURL(parameters.urlParameters)
	local resultChunks = {}
	local result, code, headers, httpStatus = http.request{
		method = "GET",
		sink = ltn12.sink.table(resultChunks),
		headers = { Cookie = parameters.httpCredentials },
		url = url
	}
	return result, code, headers, httpStatus, resultChunks
end

-- Crea un stream a partir de un archivo con el bit rate especificado
local BuildStreamFromFile = function (fileName, bitRate, bDelete)
	local file = io.open(fileName, "rb")
	if not file then
		return constants.StreamCreationResult.SOURCE_UNAVAILABLE
	end
	local stream = {
		__file = file,
		size = lfs.attributes(fileName).size,
		bitrate = bitRate,
		fileName = fileName,
		bDelete = bDelete,
		displayName = fileName:match("([^\\]*)$"),
		GetChunk = function(self, size)
			local buffer = self.__file:read(size)
			local status = (buffer and constants.StreamIOResult.MORE_DATA) or constants.StreamIOResult.LAST_CHUNK
			return buffer, status
		end,
		Release = function(self)
			self.__file:close()
			if(self.bDelete) then
				os.remove(self.fileName)
			end
		end
	}
	return stream
end

-- Descarga un archivo wav a disco
-- Devuelve la ruta al archivo descargado
local DownloadWavToTemporalFile = function(sourceArray, length)
    LogInfo("streamRepository: Creating temporaly wave file.")
	local tmpFile = GetTemporalFileName(true, "WAV")
	local file = io.open(tmpFile, "wb")
	local nextKey, buffer = next(sourceArray, nil)
	while (buffer) do
		file:write(buffer)
		nextKey, buffer = next(sourceArray, nextKey)
	end
	file:close()
	return tmpFile;
end

-- Convierte un archivo wav local a mp3 usando ffmpeg
-- Devuelve la ruta al mp3 final
local ConvertWavToMP3 = function(sourceFilename, nBitRate)
	LogInfo("streamRepository: Converting wav file to mp3.")
	local mp3FileName = GetTemporalFileName(true, "CALL")
	local cmd = string.format( [[%s -i "%s" -b %d "%s"]],
		inconcert.environment.apps.streaming.custom.ffmpegPath,
		sourceFilename,
		nBitRate * 1000,
		mp3FileName
    )   
	local exitCode = os.execute(cmd)
	if(exitCode == 0) then
		return true, mp3FileName
	else
		LogError("streamRepository: Error %d converting wav file to mp3.", exitCode)
		return false, exitCode
	end   
end


local BuildStreamWave = function(sourceArray, length)
	local tmpFile = DownloadWavToTemporalFile(sourceArray, length)
	local nBitRate = inconcert.environment.apps.streaming.custom.ffmpegBitRate
	local result, mp3FileName = ConvertWavToMP3(tmpFile, nBitRate)
	os.remove(tmpFile)
	if not result then
		return constants.StreamCreationResult.CANNOT_TRANSFORM_DATA
	end
    return BuildStreamFromFile(mp3FileName, nBitRate, true)
end


--[=[
Enva los datos que se encuentran en un array. No necesito usar un archivo temporal. 
]=]
local BuildStreamFromArray = function(sourceArray, length)
	if not sourceArray then
		return constants.StreamCreationResult.SOURCE_UNAVAILABLE
	end
	
	local SendOutput = function(stream, data, size, result)
		if data then
			if #data == size then	-- optimizo el caso en que tengo los datos justos
				stream.__remainder = nil
				return data, result
			end
			stream.__remainder = string.sub(data, size + 1)
			return string.sub(data, 1, size), result
		else
			return "", result
		end
	end
	
	local stream = {
		__source = sourceArray,
		__nextKey = nil,
		size = length,
		bitrate = 192,
		displayName = "blah!",
		__remainder = nil, -- ac me guardo temporalmente datos entre invocaciones a GetChunk
		GetChunk = function(self, size)
			-- tengo que juntar 'size' bytes del array
			local length = 0
			local buffer
			local collector = {}
			-- si me quedaron datos de la tanda anterior, los agrego ahora
			if self.__remainder then
				table.insert(collector, self.__remainder)
				length = length + #self.__remainder
				self.__remainder = nil
			end
			while(length < size) do
				self.__nextKey, buffer = next(self.__source, self.__nextKey)
				if buffer then
					table.insert(collector, buffer)
					length = length + #buffer
				else
					return SendOutput(self, table.concat(collector), size, constants.StreamIOResult.LAST_CHUNK)
				end
			end
			return SendOutput(self, table.concat(collector), size, constants.StreamIOResult.MORE_DATA)
		end,
		Release = function(self)
			-- no es necesario hacer nada
		end
	}
	
	return stream
end

--////////////////////////////////////////////////////////////////////////////////////////////////////////
--[=[
local BuildTemporalFile = function (resultChunks)
	local fileName = GetTemporalFileName(true)
	local blobFile = table.concat(resultChunks)
	local file, err = io.open(fileName, "wb")
	local result = file:write(blobFile)
	file:close()
end
--]=]

--////////////////////////////////////////////////////////////////////////////////////////////////////////
--[=[
local CreateDirectory = function (directoryName)
	assert(type(directoryName) == "string")
	os.execute("md " .. directoryName)
end
]=]

--////////////////////////////////////////////////////////////////////////////////////////////////////////
-- FUNCIONES PUBLICAS  
--////////////////////////////////////////////////////////////////////////////////////////////////////////
function iterator(stream)
	if not stream then
		return
	end
	repeat
		local data, status = stream:GetChunk(8192)
		if status == constants.StreamIOResult.LAST_CHUNK then
			LogInfo("streamRepository.iterator: Se ha terminado de enviar el archivo.")
			stream:Release()
		end
		local ok = coroutine.yield(data)
		if not ok then
			LogInfo("streamRepository.iterator: El cliente ha cortado la conexin.")
			stream:Release()
			return nil, "The client dropped the connection"
		end
	until status ~= constants.StreamIOResult.MORE_DATA
end

--[=[
]=]
function Init(parameters)
	local result, code, headers, httpStatus, resultChunks = Request(parameters)
	local stream, status
	
	if (type(code) == "number") then
		if (code == 200 and resultChunks) then
			if(headers["content-type"] == "audio/x-wav") then
				stream = BuildStreamWave(resultChunks, headers["content-length"])
			else
				stream = BuildStreamFromArray(resultChunks, headers["content-length"])
			end
			if type(stream) == "table" then
				headers["Content-Type"] = "audio/mpeg"
				headers["content-length"] = stream.size
				headers["Pragma"] = "no-cache"
				headers["Cache-Control"] = "no-store, no-cache, must-revalidate, post-check=0, pre-check=0"
			else
				status = "500 error"
				headers = {
					["Content-Type"] = "text/plain"
				}
			end
		end
		if not status then
			status = string.sub(httpStatus, 10)
		end
	else
		status = "500 " .. code
		headers = {
			["Content-Type"] = "text/plain"
		}
	end
	
	return status, headers, coroutine.wrap(function() iterator(stream) end)
end
