local table = require"table"
local string = require"string"
local tinsert = table.insert
local pairs = pairs
local setmetatable = setmetatable
local rawset = rawset
local type = type
local select = select
local next = next
local print = print
local assert = assert

-- Dummy translator. Does nothing
local m_dummyTranslator = function(...)
	return ...
end

module((...), package.seeall)

local __module_location = select(2, ...)
assert(__module_location, "should be running under inconcertWebHandler")

local m_blankImageUrl = "/shared/navigation/images/blank.gif"

--[=[
La categora tiene una metatabla que oculta el atributo 'length'. Este queda guardado en la metatable y no en la categora
]=]
local function MakeCategory()
	local category = {}
	local mt = {
		length = 0
	}
	mt.__index = mt
	mt.__newindex = function(t,k,v)
		if k == "length" then
			mt[k] = v
		else
			rawset(t, k, v)
		end
	end
	setmetatable(category, mt)
	return category
end

--[=[
Obtiene las acciones pertenecientes a la categora 'navigation'
]=]
local function GetNavigation(self)
	local actions = self.permissions:GetActions("navigation")
	if not actions then return nil end
	
	local categories = {}
	local plainActions = MakeCategory()
	
	-- agrupo por categoras y guardo cuntas hay en la metatabla de la categora, para no tener que saltearme esa clave luego
	-- acciones sin categora son acciones "planas"
	for key, action in pairs(actions) do
		if not action.Category then
			plainActions[key] = action
			plainActions.length = plainActions.length + 1
		else
			local category = categories[action.Category]
			if not category then
				category = MakeCategory()
				categories[action.Category] = category
			end
			category[key] = action
			category.length = category.length + 1
		end
	end
	return categories, plainActions
end

--[=[
Genera el markup html que representa una accin
@param defaultAction Accion por defecto (opcional)
]=]
local function ParseAction(self, web, key, action, defaultAction)
	local data = action.Data
	local curentUrlPart = web.path_info
	
	curentUrlPart = curentUrlPart:match("/?(.*)")	-- remove first slash, if present
	for command, parameter in string.gmatch(data, "([^:]*):(.*)") do
		command = string.lower(command)
		if command == "view" then
			if self.viewDefinitions[parameter] then
				local link = "/view/" .. string.lower(parameter)
				local linkClass = "iconMenu_" .. string.gsub(key, "%s", "_")
				local isCurrent = false
				--print("| |- "..action.Description .."("..string.gsub(key, "%s", "_")..")")
				if curentUrlPart == link then
					isCurrent = true
				elseif action.Id == defaultAction then
					isCurrent = true
				end
				
				return {
					link = web:link("/view/" .. string.lower(parameter)),
					linkClass = "iconMenu_" .. string.gsub(key, "%s", "_"),
					description = action.Description
				}, isCurrent
			else
				error(string.format("viewDefinitions[%s] is missing", parameter))
			end
		elseif command == "url" then
			local isCurrent = false
			if curentUrlPart == parameter then
				isCurrent = true
			elseif action.Id == defaultAction then
				isCurrent = true
			end
			--[=[
			-Si la url empieza con / es absoluta, pero pertenece al host actual.
			-Si empieza con http:// es absoluta.
			-Caso contrario, se toma como relativa a la aplicacin actual.
			]=]
			local result = {
				linkClass = "iconMenu_" .. string.gsub(key, "%s", "_"),
				description = action.Description
			}
			if string.match(parameter, "^/") then
				result.link = parameter
			elseif string.match(parameter, "^http://") then
				result.link = parameter
			else
				result.link = web:link("/" .. parameter)
			end
			return result, isCurrent
		else
			error("Navigation:No navigation item generated")
		end
	end
end


--[=[
Funciones pblicas
]=]
local function GetStyles(self)
	return {
		"/shared/styles/inconcertBackground.css",
		"/shared/navigation/styles/inconcertMenu.css",
		"/shared/styles/inconcertImages.css"
	}
end

--[=[
Cuando se est procesando el request representado por 'web', obtiene el "item navegable"  actual
Esto es una pila humeante de bosta. Hay que arreglarlo!
]=]
local function GetCurrentItem(self, web)
	local categories, plainActions = GetNavigation(self)
	if not categories then
		return ""
	end
	
	-- y armo primero las categoras 
	for key, category in pairs(categories) do
		if category.length > 0 then
			for key, action in pairs(category) do
				local curentUrlPart = web.path_info
				for command, parameter in string.gmatch(action.Data, "([^:]*):(.*)") do
					command = string.lower(command)
					if command == "view" then
						if self.viewDefinitions[parameter] then
							local link = "/view/" .. string.lower(parameter)
							if curentUrlPart == link then
								-- hasta que no se arreglen los reportes!
								if self.viewDefinitions[parameter].GUI and self.viewDefinitions[parameter].GUI.GetViewTitle then
									return self.viewDefinitions[parameter].GUI.GetViewTitle()
								else
									return nil
								end
							end
						end
					elseif command == "url" then
						if curentUrlPart == parameter then
							return parameter
						end
					end
				end
			end
		else
			local _, action = next(category)
			local curentUrlPart = web.path_info
			for command, parameter in string.gmatch(action.Data, "([^:]*):(.*)") do
				command = string.lower(command)
				if command == "view" then
					if self.viewDefinitions[parameter] then
						local link = "/view/" .. string.lower(parameter)
						if curentUrlPart == link then
							return self.viewDefinitions[parameter].GUI.GetViewTitle()
						end
					end
				elseif command == "url" then
					if curentUrlPart == parameter then
						return parameter
					end
				end
			end
		end
	end
	
	return nil
end

--[=[
Genera el markup html que representa el men de acciones.
Recibe:
@render: objeto shared.render
@web: objeto web de Orbit, con el request que se est procesando actualmente
@environment: tabla con el entorno que se hace viajar a travs del render
]=]
local function Render(self, renderer, web, environment)
	-- tomo algunos valores por defecto, pero seteables externamente
	local environment = environment or {}
	local containerClass = self.containerClass or "vMenu"
	local containerId = self.containerId or "navigationMenu"
	local template = self.customTemplate or [[/shared/navigation/pages/menu.html]]
	local defaultAction = renderer.env and renderer.env.Navigation_DefaultAction    -- can be nil (most of the time)
	
	local categories, plainActions = GetNavigation(self)
	if not categories then
		return ""
	end
	-----------------------------------------------------------------------------
	-- Las siguientes funciones se utilizan para ordenar los menues,
	--   ya sean las categorias o las acciones.
	--    En s lo que hace es recibir un hash y retornar un array ordenado con las categorias o acciones.
	-----------------------------------------------------------------------------
	--[[
		Obtiene el id de la categoria busca el primer elemento dentro del hash
		ya que todos pertenecen a la misma categoria
	--]]
	local function OrderManager_GetCategoryId(actions)
		for _, action in pairs(actions) do
			return action.Category
		end
	end
	--[[
		Evalua si se debe intercambiar los elementos del array
	--]]
	local function OrderManager_CanSwap(source, target, orderBy)
		if (source[orderBy] == nil and target[orderBy] == nil) then
			return false
		elseif (source[orderBy] == nil) then
			return true
		elseif (target[orderBy] == nil) then
			return false
		end
		return source[orderBy] < target[orderBy]
	end
	--[[	
		Obtiene el primer elemento de los dos hashes a comparar para el caso de las categorias,
		,retorna el primer elemento de cada hash y manda a evaluar si se intercambian los datos
	--]]
	local function OrderManager_CanSwapCategory(source, target, orderBy)
		local firstSourceItem
		for _, v in pairs(source) do
			firstSourceItem = v
			break
		end
		local firstTargetItem
		for _, v in pairs(target) do
			firstTargetItem = v
			break
		end
		return OrderManager_CanSwap(firstSourceItem, firstTargetItem, orderBy)
	end
	--[[
		A partir de un hash crea una tabla
	--]]
	local function OrderManager_HashToTable(tableToSort)
		local tableDestination = {}
		for _, v in pairs(tableToSort) do
			table.insert(tableDestination, v)
		end
		return tableDestination
	end
	--[[
		Ordena el array, ya sea por categoria o por accion
	--]]
	local function OrderManager_Sort(tableToSort, orderBy, swapFunction)
		if not(tableToSort) then
			return tableToSort, false
		end
		local tableOrdered = OrderManager_HashToTable(tableToSort)
		table.sort(tableOrdered, function(source, target) 
			return swapFunction(source, target, orderBy)
		end)
		return tableOrdered, true
	end
	
	-- Get the translator function or use a dummy one
	local translator = (self.translator and self.translator()) or m_dummyTranslator
	
	local env = {
		_T = function(value)
			return translator(value[1])
		end,
		containerClass = containerClass,
		containerId = containerId,
		RenderCategories = function()
			local iterator = pairs
			local hasChanged
			categories, hasChanged = OrderManager_Sort(categories, 'categoryOrder', OrderManager_CanSwapCategory)
			if type(self.beforeMenuRender) == "function" then
				--Ejecuto callback de usuario si es que este esta especificado
				self.beforeMenuRender (categories, not hasChanged)
			end
			if hasChanged then
				iterator = ipairs
			end
			for categoryId, category in iterator(categories) do
				if hasChanged then
					categoryId = OrderManager_GetCategoryId(category)
				end
				if category.length > 0 then
					local linkClass = "iconMenu_" .. string.gsub(categoryId, "%s", "_")
					cosmo.yield({
						linkClass = linkClass,
						key = categoryId,
						ParseAction = function()
							local iterator = pairs
							local hasChanged
							category, hasChanged = OrderManager_Sort(category, 'actionOrder', OrderManager_CanSwap)
							if hasChanged then
								iterator = ipairs
							end
							for actionId, action in iterator(category) do
								if hasChanged then
									actionId = action.Id
								end
								local result, current = ParseAction(self, web, actionId, action, defaultAction)
								result._template = (current and 2) or 1
								cosmo.yield(result)
							end
						end
					})
				else
					local _, action = next(category)
					cosmo.yield(ParseAction(self, web, categoryId, action, defaultAction))
				end
			end
		end,
	
		RenderActions = function()
			if plainActions.length > 0 then
				cosmo.yield({
					linkClass = "iconMenu_plainActions",
					ParseAction = function()
						for key, action in pairs(plainActions) do
							cosmo.yield(ParseAction(self, web, key, action, defaultAction))
						end
					end
				})
			end
		end,
		blankImageUrl = m_blankImageUrl
	}
	return renderer:IncludeCosmo(template, env)
end



-- Se pueden pasar las siguientes propiedades
-- containerClass
-- containerId
-- customTemplate

function new(applicationName, parameters)
	assert(parameters.permissions, "You must supply a 'permissions' field")
	local newInstance = {}
	
	-- Copio explcitamente las funciones pblicas
	newInstance.Render = Render
	newInstance.GetStyles = GetStyles
	newInstance.GetCurrentItem = GetCurrentItem
	newInstance.GetNavigation = GetNavigation
	
	if type(parameters) == "table" then
		for k,v in pairs(parameters) do
			assert(not newInstance[k])
			newInstance[k] = v
		end
	end
	
	return newInstance
end
